@@ -1,6 +1,6 @@
 /* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*  GMime
- *  Copyright (C) 2000-2010 Jeffrey Stedfast
+ *  Copyright (C) 2000-2012 Jeffrey Stedfast
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public License
@@ -29,12 +29,15 @@
 
 #include "gmime-message.h"
 #include "gmime-multipart.h"
+#include "gmime-multipart-signed.h"
+#include "gmime-multipart-encrypted.h"
 #include "gmime-part.h"
 #include "gmime-utils.h"
 #include "gmime-common.h"
 #include "gmime-stream-mem.h"
 #include "gmime-table-private.h"
 #include "gmime-parse-utils.h"
+#include "gmime-events.h"
 
 
 /**
@@ -46,14 +49,7 @@
  * A #GMimeMessage represents an rfc822 message.
  **/
 
-
-typedef void (* EventCallback) (gpointer sender, gpointer args);
-
-extern void _internet_address_list_block_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data);
-extern void _internet_address_list_unblock_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data);
-extern void _internet_address_list_add_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data);
-extern void _internet_address_list_remove_event_handler (InternetAddressList *list, EventCallback callback, gpointer user_data);
-
+extern GMimeEvent *_g_mime_header_list_get_changed_event (GMimeHeaderList *headers);
 
 static void g_mime_message_class_init (GMimeMessageClass *klass);
 static void g_mime_message_init (GMimeMessage *message, GMimeMessageClass *klass);
@@ -67,6 +63,7 @@ static const char *message_get_header (GMimeObject *object, const char *header);
 static gboolean message_remove_header (GMimeObject *object, const char *header);
 static char *message_get_headers (GMimeObject *object);
 static ssize_t message_write_to_stream (GMimeObject *object, GMimeStream *stream);
+static void message_encode (GMimeObject *object, GMimeEncodingConstraint constraint);
 
 static ssize_t write_structured (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_addrspec (GMimeStream *stream, const char *name, const char *value);
@@ -74,20 +71,21 @@ static ssize_t write_received (GMimeStream *stream, const char *name, const char
 static ssize_t write_subject (GMimeStream *stream, const char *name, const char *value);
 static ssize_t write_msgid (GMimeStream *stream, const char *name, const char *value);
 
-static void to_list_changed (InternetAddressList *list, GMimeMessage *message);
-static void cc_list_changed (InternetAddressList *list, GMimeMessage *message);
-static void bcc_list_changed (InternetAddressList *list, GMimeMessage *message);
+static void to_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message);
+static void cc_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message);
+static void bcc_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message);
 
 
 static GMimeObjectClass *parent_class = NULL;
 
+
 static struct {
 	const char *name;
-	EventCallback changed_cb;
+	GMimeEventCallback changed_cb;
 } recipient_types[] = {
-	{ "To",  (EventCallback) to_list_changed  },
-	{ "Cc",  (EventCallback) cc_list_changed  },
-	{ "Bcc", (EventCallback) bcc_list_changed }
+	{ "To",  (GMimeEventCallback) to_list_changed  },
+	{ "Cc",  (GMimeEventCallback) cc_list_changed  },
+	{ "Bcc", (GMimeEventCallback) bcc_list_changed }
 };
 
 #define N_RECIPIENT_TYPES G_N_ELEMENTS (recipient_types)
@@ -147,6 +145,14 @@ g_mime_message_class_init (GMimeMessageClass *klass)
 	object_class->get_header = message_get_header;
 	object_class->get_headers = message_get_headers;
 	object_class->write_to_stream = message_write_to_stream;
+	object_class->encode = message_encode;
+}
+
+static void
+mime_part_headers_changed (GMimeHeaderList *headers, gpointer args, GMimeMessage *message)
+{
+	/* clear the message's header stream */
+	g_mime_header_list_set_stream (((GMimeObject *) message)->headers, NULL);
 }
 
 static void
@@ -156,7 +162,7 @@ connect_changed_event (GMimeMessage *message, GMimeRecipientType type)
 	
 	list = message->recipients[type];
 	
-	_internet_address_list_add_event_handler (list, recipient_types[type].changed_cb, message);
+	g_mime_event_add (list->priv, recipient_types[type].changed_cb, message);
 }
 
 static void
@@ -166,7 +172,7 @@ disconnect_changed_event (GMimeMessage *message, GMimeRecipientType type)
 	
 	list = message->recipients[type];
 	
-	_internet_address_list_remove_event_handler (list, recipient_types[type].changed_cb, message);
+	g_mime_event_remove (list->priv, recipient_types[type].changed_cb, message);
 }
 
 static void
@@ -176,7 +182,7 @@ block_changed_event (GMimeMessage *message, GMimeRecipientType type)
 	
 	list = message->recipients[type];
 	
-	_internet_address_list_block_event_handler (list, recipient_types[type].changed_cb, message);
+	g_mime_event_block (list->priv, recipient_types[type].changed_cb, message);
 }
 
 static void
@@ -186,7 +192,7 @@ unblock_changed_event (GMimeMessage *message, GMimeRecipientType type)
 	
 	list = message->recipients[type];
 	
-	_internet_address_list_unblock_event_handler (list, recipient_types[type].changed_cb, message);
+	g_mime_event_unblock (list->priv, recipient_types[type].changed_cb, message);
 }
 
 static void
@@ -232,6 +238,7 @@ static void
 g_mime_message_finalize (GObject *object)
 {
 	GMimeMessage *message = (GMimeMessage *) object;
+	GMimeEvent *changed;
 	guint i;
 	
 	g_free (message->from);
@@ -250,8 +257,11 @@ g_mime_message_finalize (GObject *object)
 	g_free (message->message_id);
 	
 	/* unref child mime part */
-	if (message->mime_part)
+	if (message->mime_part) {
+		changed = _g_mime_header_list_get_changed_event (message->mime_part->headers);
+		g_mime_event_remove (changed, (GMimeEventCallback) mime_part_headers_changed, message);
 		g_object_unref (message->mime_part);
+	}
 	
 	G_OBJECT_CLASS (parent_class)->finalize (object);
 }
@@ -781,10 +791,12 @@ message_prepend_header (GMimeObject *object, const char *header, const char *val
 {
 	GMimeMessage *message = (GMimeMessage *) object;
 	
-	/* Make sure that the header is not a Content-* header, else it
-           doesn't belong on a message */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
+	/* Content-* headers don't belong on the message, they belong on the part. */
+	if (!g_ascii_strncasecmp ("Content-", header, 8)) {
+		if (message->mime_part)
+			g_mime_object_prepend_header (message->mime_part, header, value);
 		return;
+	}
 	
 	if (!process_header (object, PREPEND, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->prepend_header (object, header, value);
@@ -800,10 +812,12 @@ message_append_header (GMimeObject *object, const char *header, const char *valu
 {
 	GMimeMessage *message = (GMimeMessage *) object;
 	
-	/* Make sure that the header is not a Content-* header, else it
-           doesn't belong on a message */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
+	/* Content-* headers don't belong on the message, they belong on the part. */
+	if (!g_ascii_strncasecmp ("Content-", header, 8)) {
+		if (message->mime_part)
+			g_mime_object_append_header (message->mime_part, header, value);
 		return;
+	}
 	
 	if (!process_header (object, APPEND, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->append_header (object, header, value);
@@ -819,10 +833,12 @@ message_set_header (GMimeObject *object, const char *header, const char *value)
 {
 	GMimeMessage *message = (GMimeMessage *) object;
 	
-	/* Make sure that the header is not a Content-* header, else it
-           doesn't belong on a message */
-	if (!g_ascii_strncasecmp ("Content-", header, 8))
+	/* Content-* headers don't belong on the message, they belong on the part. */
+	if (!g_ascii_strncasecmp ("Content-", header, 8)) {
+		if (message->mime_part)
+			g_mime_object_set_header (message->mime_part, header, value);
 		return;
+	}
 	
 	if (!process_header (object, SET, header, value))
 		GMIME_OBJECT_CLASS (parent_class)->set_header (object, header, value);
@@ -839,8 +855,7 @@ message_get_header (GMimeObject *object, const char *header)
 	GMimeMessage *message = (GMimeMessage *) object;
 	const char *value;
 	
-	/* Make sure that the header is not a Content-* header, else it
-           doesn't belong on a message */
+	/* Content-* headers don't belong on the message, they belong on the part. */
 	if (g_ascii_strncasecmp ("Content-", header, 8) != 0) {
 		if ((value = GMIME_OBJECT_CLASS (parent_class)->get_header (object, header)))
 			return value;
@@ -862,8 +877,7 @@ message_remove_header (GMimeObject *object, const char *header)
 	GMimeRecipientType type;
 	guint i;
 	
-	/* Make sure that the header is not a Content-* header, else it
-           doesn't belong on a message */
+	/* Content-* headers don't belong on the message, they belong on the part. */
 	if (!g_ascii_strncasecmp ("Content-", header, 8)) {
 		if (message->mime_part)
 			return g_mime_object_remove_header (message->mime_part, header);
@@ -945,11 +959,6 @@ message_get_headers (GMimeObject *object)
 		/* if the mime part has raw headers, then it contains the message headers as well */
 		g_mime_header_list_write_to_stream (message->mime_part->headers, stream);
 	} else {
-		/* Note: if the user changed the mime_part's headers,
-		 * we need to unset our cached header stream or it
-		 * won't reflect the changes. */
-		g_mime_header_list_set_stream (object->headers, NULL);
-		
 		g_mime_header_list_write_to_stream (object->headers, stream);
 		if (message->mime_part) {
 			if (g_mime_object_get_header (message->mime_part, "Content-Type") &&
@@ -975,11 +984,6 @@ message_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	
 	if (message->mime_part) {
 		if (!g_mime_header_list_get_stream (message->mime_part->headers)) {
-			/* Note: if the user changed the mime_part's headers,
-			 * we need to unset our cached header stream or it
-			 * won't reflect the changes. */
-			g_mime_header_list_set_stream (object->headers, NULL);
-			
 			if ((nwritten = g_mime_header_list_write_to_stream (object->headers, stream)) == -1)
 				return -1;
 			
@@ -1010,9 +1014,19 @@ message_write_to_stream (GMimeObject *object, GMimeStream *stream)
 	return total;
 }
 
+static void
+message_encode (GMimeObject *object, GMimeEncodingConstraint constraint)
+{
+	GMimeMessage *message = (GMimeMessage *) object;
+	
+	if (message->mime_part != NULL)
+		g_mime_object_encode (message->mime_part, constraint);
+}
+
+
 
 /**
- * g_mime_message_new: Create a new MIME Message object
+ * g_mime_message_new:
  * @pretty_headers: make pretty headers 
  *
  * If @pretty_headers is %TRUE, then the standard rfc822 headers are
@@ -1020,7 +1034,7 @@ message_write_to_stream (GMimeObject *object, GMimeStream *stream)
  * strictly a cosmetic thing, so if you are unsure, it is safe to say
  * no (%FALSE).
  *
- * Returns: an empty MIME Message object.
+ * Returns: an empty #GMimeMessage object.
  **/
 GMimeMessage *
 g_mime_message_new (gboolean pretty_headers)
@@ -1071,9 +1085,6 @@ g_mime_message_set_sender (GMimeMessage *message, const char *sender)
 		g_mime_header_list_set (GMIME_OBJECT (message)->headers, "From", "");
 		message->from = NULL;
 	}
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1111,9 +1122,6 @@ g_mime_message_set_reply_to (GMimeMessage *message, const char *reply_to)
 	message->reply_to = g_mime_strdup_trim (reply_to);
 	
 	g_mime_header_list_set (GMIME_OBJECT (message)->headers, "Reply-To", message->reply_to);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1151,25 +1159,22 @@ sync_recipient_header (GMimeMessage *message, GMimeRecipientType type)
 		/* list should never be NULL... */
 		g_mime_header_list_set (object->headers, name, NULL);
 	}
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 static void
-to_list_changed (InternetAddressList *list, GMimeMessage *message)
+to_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message)
 {
 	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_TO);
 }
 
 static void
-cc_list_changed (InternetAddressList *list, GMimeMessage *message)
+cc_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message)
 {
 	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_CC);
 }
 
 static void
-bcc_list_changed (InternetAddressList *list, GMimeMessage *message)
+bcc_list_changed (InternetAddressList *list, gpointer args, GMimeMessage *message)
 {
 	sync_recipient_header (message, GMIME_RECIPIENT_TYPE_BCC);
 }
@@ -1198,11 +1203,6 @@ g_mime_message_add_recipient (GMimeMessage *message, GMimeRecipientType type, co
 	ia = internet_address_mailbox_new (name, addr);
 	internet_address_list_add (recipients, ia);
 	g_object_unref (ia);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
-	
-	g_mime_header_list_set_stream (((GMimeObject *) message)->headers, NULL);
 }
 
 
@@ -1280,9 +1280,6 @@ g_mime_message_set_subject (GMimeMessage *message, const char *subject)
 	encoded = g_mime_utils_header_encode_text (message->subject);
 	g_mime_object_set_header (GMIME_OBJECT (message), "Subject", encoded);
 	g_free (encoded);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1324,9 +1321,6 @@ g_mime_message_set_date (GMimeMessage *message, time_t date, int tz_offset)
 	str = g_mime_utils_header_format_date (date, tz_offset);
 	g_mime_object_set_header (GMIME_OBJECT (message), "Date", str);
 	g_free (str);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1393,9 +1387,6 @@ g_mime_message_set_date_as_string (GMimeMessage *message, const char *str)
 	buf = g_mime_utils_header_format_date (date, tz_offset);
 	g_mime_object_set_header (GMIME_OBJECT (message), "Date", buf);
 	g_free (buf);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1420,9 +1411,6 @@ g_mime_message_set_message_id (GMimeMessage *message, const char *message_id)
 	msgid = g_strdup_printf ("<%s>", message_id);
 	g_mime_object_set_header (GMIME_OBJECT (message), "Message-Id", msgid);
 	g_free (msgid);
-	
-	if (message->mime_part)
-		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 }
 
 
@@ -1473,17 +1461,29 @@ g_mime_message_get_mime_part (GMimeMessage *message)
 void
 g_mime_message_set_mime_part (GMimeMessage *message, GMimeObject *mime_part)
 {
-	g_return_if_fail (GMIME_IS_MESSAGE (message));
-	g_return_if_fail (GMIME_IS_OBJECT (mime_part));
+	GMimeEvent *changed;
 	
-	g_object_ref (mime_part);
-	g_mime_header_list_set_stream (mime_part->headers, NULL);
+	g_return_if_fail (mime_part == NULL || GMIME_IS_OBJECT (mime_part));
+	g_return_if_fail (GMIME_IS_MESSAGE (message));
+	
+	if (message->mime_part == mime_part)
+		return;
 	
 	if (message->mime_part) {
+		changed = _g_mime_header_list_get_changed_event (message->mime_part->headers);
+		g_mime_event_remove (changed, (GMimeEventCallback) mime_part_headers_changed, message);
+		
 		g_mime_header_list_set_stream (message->mime_part->headers, NULL);
 		g_object_unref (message->mime_part);
 	}
 	
+	if (mime_part) {
+		changed = _g_mime_header_list_get_changed_event (mime_part->headers);
+		g_mime_header_list_set_stream (mime_part->headers, NULL);
+		g_mime_event_add (changed, (GMimeEventCallback) mime_part_headers_changed, message);
+		g_object_ref (mime_part);
+	}
+	
 	g_mime_header_list_set_stream (((GMimeObject *) message)->headers, NULL);
 	
 	message->mime_part = mime_part;
@@ -1509,3 +1509,88 @@ g_mime_message_foreach (GMimeMessage *message, GMimeObjectForeachFunc callback,
 	if (GMIME_IS_MULTIPART (message->mime_part))
 		g_mime_multipart_foreach ((GMimeMultipart *) message->mime_part, callback, user_data);
 }
+
+static gboolean
+part_is_textual (GMimeObject *mime_part)
+{
+	GMimeContentType *type;
+	
+	type = g_mime_object_get_content_type (mime_part);
+	
+	return g_mime_content_type_is_type (type, "text", "*");
+}
+
+static GMimeObject *
+multipart_guess_body (GMimeMultipart *multipart)
+{
+	GMimeContentType *type;
+	GMimeObject *mime_part;
+	int count, i;
+	
+	if (GMIME_IS_MULTIPART_ENCRYPTED (multipart)) {
+		/* nothing more we can do */
+		return (GMimeObject *) multipart;
+	}
+	
+	type = g_mime_object_get_content_type ((GMimeObject *) multipart);
+	if (g_mime_content_type_is_type (type, "multipart", "alternative")) {
+		/* very likely that this is the body - leave it up to
+		 * our caller to decide which format of the body it
+		 * wants to use. */
+		return (GMimeObject *) multipart;
+	}
+	
+	count = g_mime_multipart_get_count (multipart);
+	
+	if (count >= 1 && GMIME_IS_MULTIPART_SIGNED (multipart)) {
+		/* if the body is in here, it has to be the first part */
+		count = 1;
+	}
+	
+	for (i = 0; i < count; i++) {
+		mime_part = g_mime_multipart_get_part (multipart, i);
+		
+		if (GMIME_IS_MULTIPART (mime_part)) {
+			if ((mime_part = multipart_guess_body ((GMimeMultipart *) mime_part)))
+				return mime_part;
+		} else if (GMIME_IS_PART (mime_part)) {
+			if (part_is_textual (mime_part))
+				return mime_part;
+		}
+	}
+	
+	return NULL;
+}
+
+
+/**
+ * g_mime_message_get_body:
+ * @message: MIME Message
+ *
+ * Attempts to identify the MIME part containing the body of the
+ * message.
+ *
+ * Returns: a #GMimeObject containing the textual content that appears
+ * to be the main body of the message.
+ *
+ * Note: This function is NOT guarenteed to always work as it
+ * makes some assumptions that are not necessarily true. It is
+ * recommended that you traverse the MIME structure yourself.
+ **/
+GMimeObject *
+g_mime_message_get_body (GMimeMessage *message)
+{
+	GMimeObject *mime_part;
+	
+	g_return_val_if_fail (GMIME_IS_MESSAGE (message), NULL);
+	
+	if (!(mime_part = message->mime_part))
+		return NULL;
+	
+	if (GMIME_IS_MULTIPART (mime_part))
+		return multipart_guess_body ((GMimeMultipart *) mime_part);
+	else if (GMIME_IS_PART (mime_part) && part_is_textual (mime_part))
+		return mime_part;
+	
+	return NULL;
+}
